/* hammerhead.c

	creates a /proc entry to get details
	on the hammerhead controller
	and the physical memory configuration

	quick&dirty hack by M. Lanners <mlan@cpu.lu>
	
	based on code by Apple Computer, Inc.

03-10-2002:
	Andreas Tobler <a.tobler@schweiz.ch> 	

	- Update proc_fs calls
	- Detect Physical Memory
	- Detect Machine

15-12-2002:
	Andreas Tobler <a.tobler@schweiz.ch>
	- added ANS detection, thanks to eddied@xs4all.nl
 
29-12-2002:
	Eddie Bindt <eddieb@xs4all.nl>
	- ANS has same bank layout as 9500 but only 8 slots
*/

#ifndef MODULE
# define MODULE
#endif

#ifndef __KERNEL__
# define __KERNEL__
#endif

#define __NO_VERSION__
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/version.h>

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/kmod.h>

#include <asm/io.h>
#include <asm/prom.h>
#include <asm/machdep.h>
#include <asm/pmac_feature.h>

#define interleaveMask 		0x04
#define adrs1Mask		0x01
#define adrs1Shift		8 
#define MegPerAdrs		4 

static char			driver_version[] = "0.4";

/* ANS */
int	ans	=	0;

struct RAMBankDescriptor {
        int			bankSize;
        int			bankIsInterleaved;
};

typedef struct RAMBankDescriptor RAMBankDescriptor;
typedef RAMBankDescriptor* RAMBankDescriptorPtr;

struct MacModel {
	int 		Model;	
	char*		Name;
};

typedef struct MacModel MacModel;
typedef MacModel* MacModelPtr;
void GetPhysicalRAMBankInfo(RAMBankDescriptorPtr  theBanks, unsigned long  fullMemSize, unsigned long hammerHeadBase)
{
        char *		bankN;
        short		i;
        
        bankN = (char *)hammerHeadBase + 0x1C0;
    
        for (i=0; i<=25; i++)
        {
                theBanks[i].bankIsInterleaved = (*bankN) & interleaveMask;
                theBanks[i].bankSize = (*bankN) & adrs1Mask;
                bankN = bankN + 0x10;
                theBanks[i].bankSize = (theBanks[i].bankSize << adrs1Shift) + (*
bankN);
                bankN = bankN + 0x10;
        } 
        
        for (i = 0; i < 25; i++)
        {
                theBanks[i].bankSize = (theBanks[i+1].bankSize - theBanks[i].bankSize) * MegPerAdrs;
        }

        theBanks[25].bankSize = (fullMemSize / 4 - theBanks[25].bankSize) * MegPerAdrs;

        for (i = 0; i < 25; i++)
        {
                if (theBanks[i].bankIsInterleaved) {
                        theBanks[i+1].bankSize = theBanks[i].bankSize / 2;
                        theBanks[i].bankSize = theBanks[i+1].bankSize;
                }
	} 
        
        return;
}

static void
PrintBankInfo(int k, RAMBankDescriptorPtr theBanks, char *buf, int *len)
{
        (*len)+= sprintf (buf+(*len), "%3.0d/%3.0d ", theBanks[k].bankSize, theBanks[k+1].bankSize);

        if (theBanks[k].bankIsInterleaved)
                (*len)+= sprintf (buf+(*len), "Interleaved");
        else
                (*len)+= sprintf (buf+(*len), " --- ");

        (*len)+= sprintf(buf+(*len), "\n");
}

static void
Print75008500 (RAMBankDescriptorPtr  theBanks, char *buf, int *len)
{
	int	i;
	int	j;

	(*len)+= sprintf (buf+(*len), "Motherboard Row-> not implemented\n"); 

        for (i=1; i<=4; i++)    
        {
                (*len)+= sprintf (buf+(*len), "A%d/B%d (row 0) -> ", i, i);   
                j = 18-(4*i);
                PrintBankInfo(j, theBanks, buf, len);
        
                (*len)+= sprintf (buf+(*len), "A%d/B%d (row 1) -> ", i, i);   
                j = 20-(4*i);
                PrintBankInfo(j, theBanks, buf, len);
        }
}

static void 
Print9500 (RAMBankDescriptorPtr  theBanks, char *buf, int *len)
{
	int	i, j, k;
	k = 6;			/* number of rows to print */
	if (ans) k = 4;		/* ANS just has 8 rows     */

	(*len)+= sprintf (buf+(*len), "Motherboard Row-> not implemented\n"); 

        /* Decipher the bank information, and display it. */
        for (i=1; i<=k; i++)            
        {
                (*len)+= sprintf (buf+(*len), "A%d/B%d (row 0) -> ", i, i);   
                j = (4*i)-2;
                PrintBankInfo(j, theBanks, buf, len);
                
                (*len)+= sprintf (buf+(*len), "A%d/B%d (row 1) -> ", i, i);   
                j = 4*i;
                PrintBankInfo(j, theBanks, buf, len);
        }
}               

static void
PrintGenericMac (RAMBankDescriptorPtr  theBanks, char *buf, int *len)
{
	int	i, j;

	for (i = 0; i < 13; i++)
	{
		(*len)+= sprintf (buf+(*len), "x%d/x%d         -> ", i, i);
		j = 2 * i;
		PrintBankInfo(j, theBanks, buf, len);
	}
}

unsigned long findhammerheadbase(void)
{
	struct  device_node *dev;
	unsigned long 	addr;
	
	dev = find_devices("hammerhead");
	if (dev == 0)
		return 0;
	
	addr = (long)ioremap (dev->addrs[0].address, 0x400);
	return addr;
}

int GetPhysicalRamSize(void)
{
        struct  device_node *np;
	int	phyRamSize, n;
	unsigned long total = 0;
	struct reg_property *reg;

        /* find ram info */
        np = find_devices("memory");
        if (np == 0)
		return 0;

	reg = (struct reg_property *)
	get_property(np, "reg", &n);

	if (reg == 0) 
		return 0;

	for (n /= sizeof(struct reg_property); n > 0; --n)
		total += (reg++)->size;
		phyRamSize = total >> 20;

	return phyRamSize;
}

void GetMacModel(MacModelPtr macModel)
{
	int mbmodel;
	char* mbname;

	mbmodel = pmac_call_feature(PMAC_FTR_GET_MB_INFO,
		NULL, PMAC_MB_INFO_MODEL, 0);

        if (pmac_call_feature(PMAC_FTR_GET_MB_INFO, NULL, PMAC_MB_INFO_NAME,
		(int)&mbname) != 0)
		mbname = "Unknown";
	
	macModel->Model = mbmodel;
	macModel->Name = mbname;
}

static int hammerhead_read_procmem(char *buf, char **start, off_t offset, int len)
{
	unsigned long		hammerheadbase;
	int			RAMSize;
	int			type;
	RAMBankDescriptor	allBanks[26];
	MacModel		mac_model;

	len = 0;
	type = 0;
	
        GetMacModel(&mac_model);

        len += sprintf (buf+len,"Machine detected = %s \n", mac_model.Name);
        len += sprintf (buf+len,"Model Type = %d \n", mac_model.Model);
	if (strcmp(mac_model.Name, "PowerMac 7500") == 0)
		type = 7500;
	if (strcmp(mac_model.Name, "PowerMac 8500/8600") == 0)
                type = 7500;
	if (strcmp(mac_model.Name, "PowerMac 9500/9600") == 0)
                type = 9500; 
	if (strcmp(mac_model.Name, "Apple Network Server") == 0)
	{
		type = 9500;
		ans = 1;
	};

	if ((hammerheadbase = findhammerheadbase()) == 0) {
		len += sprintf (buf+len, "---Unsupported Macintosh!---\n");
		return len;
	}

	if ((RAMSize = GetPhysicalRamSize()) == 0) {
		len += sprintf (buf+len,"---Memory Size unspecified!---\n");
		return len;
	}
	else
		len += sprintf (buf+len,"Total available Memory = %d MB in slot config:\n",
			RAMSize);

	GetPhysicalRAMBankInfo(allBanks, RAMSize, hammerheadbase);
	

	switch (type)
	{
	  case 7500:
		Print75008500(allBanks, buf, &len);
		break;
	  case 9500:
		Print9500(allBanks, buf, &len);
		break;
	  default:
		len += sprintf (buf+len,"Unknown PowerMac: Meg/Meg Interleaved?\n");
		PrintGenericMac(allBanks, buf, &len);
		break;
	}

	return len;
}

/***********************************
 * General module initialization   *
 ***********************************/

static int __init hammerhead_init(void)
{
	struct proc_dir_entry *proc_hammerhead_entry;

        proc_hammerhead_entry = create_proc_info_entry("hammerhead",
		0, NULL, hammerhead_read_procmem);

	if (!proc_hammerhead_entry)
		return -ENOENT;
	
        if (proc_hammerhead_entry)
                SET_MODULE_OWNER(proc_hammerhead_entry);

	printk(KERN_INFO "Hammerhead memory display %s initialized.\n",
		driver_version);

	return 0;
}

static void __exit hammerhead_cleanup(void)
{
	/* Remove the proc entry */
	remove_proc_entry("hammerhead", NULL);

        printk(KERN_INFO "Hammerhead removed.\n");
}

module_init(hammerhead_init);
module_exit(hammerhead_cleanup);

MODULE_AUTHOR("Michel Lanners <mlan@cpu.lu>");
MODULE_DESCRIPTION("Display the memory config of the Hammerhead Controller");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
